'use strict';

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var Promise,
    Runner,
    checkResults,
    common,
    debug,
    debugReceivedMessages,
    expectation,
    fbp,
    fbpClient,
    getFixtureGraph,
    loadComponentSuites,
    protocol,
    runAll,
    runSuite,
    runTestAndCheck,
    sendMessageAndWait,
    synthesizeTopicFixture,
    testsuite,
    indexOf = [].indexOf;

common = require('./common');

protocol = require('./protocol');

testsuite = require('./testsuite');

expectation = require('./expectation');

fbp = require('fbp');

fbpClient = require('fbp-protocol-client');

debug = require('debug')('fbp-spec:runner');

Promise = require('bluebird');

debugReceivedMessages = function debugReceivedMessages(client) {
  var debugReceived;
  debugReceived = require('debug')('fbp-spec:runner:received');
  client.on('graph', function (_ref) {
    var command = _ref.command,
        payload = _ref.payload;

    return debugReceived('graph', command, payload);
  });
  client.on('network', function (_ref2) {
    var command = _ref2.command,
        payload = _ref2.payload;

    return debugReceived('network', command, payload);
  });
  client.on('runtime', function (_ref3) {
    var command = _ref3.command,
        payload = _ref3.payload;

    return debugReceived('runtime', command, payload);
  });
  client.on('component', function (_ref4) {
    var command = _ref4.command,
        payload = _ref4.payload;

    return debugReceived('component', command, payload);
  });
  return client.on('execution', function (status) {
    return debugReceived('execution', status);
  });
};

synthesizeTopicFixture = function synthesizeTopicFixture(topic, components, callback) {
  var component, graph, j, k, len, len1, port, portName, processName, ref, ref1;
  debug('synthesizing fixture', topic);
  // export each of the ports for topic component
  // return a FBP graph?
  component = components[topic];
  if (!component) {
    return callback(new Error('Could not find component for topic: ' + topic));
  }
  graph = {
    properties: {},
    inports: {},
    outports: {},
    processes: {},
    connections: []
  };
  processName = 'testee';
  graph.processes[processName] = {
    component: topic
  };
  ref = component.inPorts;
  for (j = 0, len = ref.length; j < len; j++) {
    port = ref[j];
    portName = port.id;
    graph.inports[portName] = {
      process: processName,
      port: portName
    };
  }
  ref1 = component.outPorts;
  for (k = 0, len1 = ref1.length; k < len1; k++) {
    port = ref1[k];
    portName = port.id;
    graph.outports[portName] = {
      process: processName,
      port: portName
    };
  }
  // Sanity checking if this is usable as a fixture
  if (Object.keys(graph.outports) < 1) {
    return callback(new Error('Component \'' + topic + '\' used as fixture has no outports'));
  }
  if (Object.keys(graph.inports) < 1) {
    return callback(new Error('Component \'' + topic + '\' used as fixture has no inports'));
  }
  return callback(null, graph);
};

// @context should have .components = {} property
getFixtureGraph = function getFixtureGraph(context, suite, callback) {
  var hasComponents, updateComponents;
  // TODO: follow runtime for component changes
  hasComponents = function hasComponents(s) {
    return s.components != null && _typeof(s.components) === 'object' && Object.keys(s.components).length;
  };
  updateComponents = function updateComponents(cb) {
    if (hasComponents(context)) {
      return cb(null);
    }
    return protocol.getComponents(context.client, function (err, components) {
      if (err) {
        return cb(err);
      }
      context.components = components;
      return cb(null);
    });
  };
  return updateComponents(function (err) {
    var e, graph;
    if (err) {
      return callback(err);
    }
    if (suite.fixture == null) {
      return synthesizeTopicFixture(suite.topic, context.components, callback);
    } else if (suite.fixture.type === 'json') {
      try {
        graph = JSON.parse(suite.fixture.data);
      } catch (error1) {
        e = error1;
        return callback(e);
      }
      return callback(null, graph);
    } else if (suite.fixture.type === 'fbp') {
      try {
        graph = fbp.parse(suite.fixture.data);
      } catch (error1) {
        return callback(e);
      }
      if (!graph.properties) {
        graph.properties = {};
      }
      return callback(null, graph);
    } else {
      return callback(new Error('Unknown fixture type ' + suite.fixture.type));
    }
  });
};

sendMessageAndWait = function sendMessageAndWait(client, currentGraph, inputData, expectData, callback) {
  var checkPacket, onReceived, received;
  received = {};
  onReceived = function onReceived(port, data) {
    var nExpected;
    debug('runtest got output on', port);
    received[port] = data;
    nExpected = Object.keys(expectData).length;
    if (Object.keys(received).length === nExpected) {
      client.removeListener('runtime', checkPacket);
      return callback(null, received);
    }
  };
  checkPacket = function checkPacket(msg) {
    var d;
    d = msg.payload;
    // FIXME: also check # and d.graph == @currentGraphId
    if (msg.command === 'packet' && d.event === 'data') {
      return onReceived(d.port, d.payload);
    } else if (msg.command === 'packet' && ['begingroup', 'endgroup', 'connect', 'disconnect'].indexOf(d.event) !== -1) {} else {
      // ignored
      return debug('unknown runtime message', msg);
    }
  };
  client.on('runtime', checkPacket);
  // send input packets
  return protocol.sendPackets(client, currentGraph, inputData, function (err) {
    if (err) {
      return callback(err);
    }
  });
};

Runner = function () {
  function Runner(client1) {
    var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    _classCallCheck(this, Runner);

    var Transport;
    this.client = client1;
    if (this.client.protocol != null && this.client.address != null) {
      // is a runtime definition
      Transport = fbpClient.getTransport(this.client.protocol);
      this.client = new Transport(this.client);
    }
    this.currentGraphId = null;
    this.components = {};
    this.options = options;
    if (this.options.connectTimeout == null) {
      this.options.connectTimeout = 5 * 1000;
    }
  }

  // TODO: check the runtime capabilities before continuing


  _createClass(Runner, [{
    key: 'connect',
    value: function connect(callback) {
      var _this = this;

      var attempts, isOnline, timeBetweenAttempts, tryConnect;
      debug('connect');
      debugReceivedMessages(this.client);
      this.client.on('network', function (_ref5) {
        var command = _ref5.command,
            payload = _ref5.payload;

        if (command === 'output' && payload.message) {
          return console.log(payload.message);
        }
      });
      this.client.on('error', function (err) {
        return debug('connection failed', err);
      });
      timeBetweenAttempts = 500;
      attempts = Math.floor(this.options.connectTimeout / timeBetweenAttempts);
      isOnline = function isOnline() {
        var connected;
        connected = _this.client.isConnected();
        if (connected) {
          return Promise.resolve();
        } else {
          return Promise.reject(new Error('Not connected to runtime'));
        }
      };
      tryConnect = function tryConnect() {
        debug('trying to connect');
        _this.client.connect(); // does not always emit an event, so we don't bother checking any
        return Promise.resolve();
      };
      return common.retryUntil(tryConnect, isOnline, timeBetweenAttempts, attempts).asCallback(callback);
    }
  }, {
    key: 'disconnect',
    value: function disconnect(callback) {
      var _this2 = this;

      var _onStatus;
      debug('disconnect');
      _onStatus = function onStatus(status) {
        var err;
        err = !status.online ? null : new Error('Runtime online after disconnect()');
        _this2.client.removeListener('status', _onStatus);
        debug('disconnected', err);
        return callback(err);
      };
      this.client.on('status', _onStatus);
      return this.client.disconnect();
    }
  }, {
    key: 'setupSuite',
    value: function setupSuite(suite, callback) {
      var _this3 = this;

      debug('setup suite', '"' + suite.name + '"');
      return getFixtureGraph(this, suite, function (err, graph) {
        if (err) {
          return callback(err);
        }
        return protocol.sendGraph(_this3.client, graph, function (err, graphId) {
          _this3.currentGraphId = graphId;
          if (err) {
            return callback(err);
          }
          return protocol.startNetwork(_this3.client, graphId, function (err) {
            return callback(err);
          });
        });
      });
    }
  }, {
    key: 'teardownSuite',
    value: function teardownSuite(suite, callback) {
      debug('teardown suite', '"' + suite.name + '"');
      // FIXME: also remove the graph. Ideally using a 'destroy' message in FBP protocol
      return protocol.stopNetwork(this.client, this.currentGraphId, function (err) {
        return callback(err);
      });
    }
  }, {
    key: 'runTest',
    value: function runTest(testcase, callback) {
      var _this4 = this;

      var expects, i, inputs, j, ref, sendWait, sequence;
      debug('runtest', '"' + testcase.name + '"');
      // XXX: normalize and validate in testsuite.coffee instead?
      inputs = common.isArray(testcase.inputs) ? testcase.inputs : [testcase.inputs];
      expects = common.isArray(testcase.expect) ? testcase.expect : [testcase.expect];
      sequence = [];
      for (i = j = 0, ref = inputs.length; 0 <= ref ? j < ref : j > ref; i = 0 <= ref ? ++j : --j) {
        sequence.push({
          inputs: inputs[i],
          expect: expects[i]
        });
      }
      sendWait = function sendWait(data, cb) {
        return sendMessageAndWait(_this4.client, _this4.currentGraphId, data.inputs, data.expect, cb);
      };
      return common.asyncSeries(sequence, sendWait, function (err, actuals) {
        actuals.forEach(function (r, idx) {
          return sequence[idx].actual = r;
        });
        return callback(err, sequence);
      });
    }
  }]);

  return Runner;
}();

// TODO: should this go into expectation?
checkResults = function checkResults(results) {
  var actuals, errors, expects, failed, failures, index, result;
  actuals = results.filter(function (r) {
    return r.actual != null;
  });
  expects = results.filter(function (r) {
    return r.expect != null;
  });
  if (actuals.length < expects.length) {
    return callback(null, {
      passed: false,
      error: new Error('Only got ' + actual.length + ' output messages out of ' + expect.length)
    });
  }
  results = results.map(function (res) {
    var e;
    res.error = null;
    try {
      expectation.expect(res.expect, res.actual);
    } catch (error1) {
      e = error1;
      // FIXME: only catch actual AssertionErrors
      res.error = e;
    }
    return res;
  });
  failures = results.filter(function (r) {
    return r.error;
  });
  if (failures.length === 0) {
    result = {
      passed: true
    };
  } else {
    if (expects.length === 1) {
      result = {
        error: failures[0].error
      };
    } else if (expects.length > 1 && failures.length === 1) {
      index = results.findIndex(function (r) {
        return r.error;
      });
      failed = results[index];
      result = {
        error: new Error('Expectation ' + index + ' of sequence failed: ' + failed.error.message)
      };
    } else {
      errors = results.map(function (r) {
        var ref;
        return ((ref = r.error) != null ? ref.message : void 0) || '';
      });
      result = {
        error: new Error('Multiple failures in sequence: ' + errors)
      };
    }
  }
  return result;
};

runTestAndCheck = function runTestAndCheck(runner, testcase, callback) {
  var expects, inputs;
  if (testcase.skip) {
    return callback(null, {
      passed: true
    });
  }
  // TODO: pass some skipped state? its indirectly in .skip though

  // XXX: normalize and validate in testsuite.coffee instead?
  inputs = common.isArray(testcase.inputs) ? testcase.inputs : [testcase.inputs];
  expects = common.isArray(testcase.expect) ? testcase.expect : [testcase.expect];
  if (inputs.length !== expects.length) {
    return callback(null, {
      passed: false,
      error: new Error('Test sequence length mismatch. Got ' + inputs.length + ' inputs and ' + expects.length + ' expectations')
    });
  }
  return runner.runTest(testcase, function (err, results) {
    var result;
    if (err) {
      return callback(err, null);
    }
    result = checkResults(results);
    if (result.error) {
      result.passed = false;
    }
    return callback(null, result);
  });
};

runSuite = function runSuite(runner, suite, runTest, callback) {
  return runner.setupSuite(suite, function (err) {
    debug('setup suite', err);
    if (err) {
      return callback(err, suite);
    }
    return common.asyncSeries(suite.cases, runTest, function (err) {
      debug('testrun complete', err);
      return runner.teardownSuite(suite, function (err) {
        debug('teardown suite', err);
        return callback(err, suite);
      });
    });
  });
};

exports.getComponentSuites = function (runner, callback) {
  return protocol.getCapabilities(runner.client, function (err, caps) {
    if (err) {
      return callback(err);
    }
    if (indexOf.call(caps, 'component:getsource') < 0) {
      return callback(null, []);
    }
    return protocol.getComponentTests(runner.client, function (err, tests) {
      var suites;
      if (err) {
        return callback(err);
      }
      suites = loadComponentSuites(tests);
      debug('get component suites', tests.length, suites.length);
      return callback(null, suites);
    });
  });
};

loadComponentSuites = function loadComponentSuites(componentTests) {
  var e, name, ss, suites, tests;
  suites = [];
  for (name in componentTests) {
    tests = componentTests[name];
    try {
      ss = testsuite.loadYAML(tests);
      suites = suites.concat(ss);
    } catch (error1) {
      e = error1;
      // ignore, could be non fbp-spec test
      // TODO: include tests type in FBP protocol, so we know whether this is error or legit
      continue;
    }
  }
  return suites;
};

// will update each of the testcases in @suites
// with .passed and .error states as tests are ran
runAll = function runAll(runner, suites, updateCallback, doneCallback) {
  var runOneSuite, runTest, s;
  runTest = function runTest(testcase, callback) {
    var done;
    done = function done(error) {
      updateCallback(suites);
      return callback(error);
    };
    return runTestAndCheck(runner, testcase, function (err, results) {
      var key, val;
      for (key in results) {
        val = results[key];
        testcase[key] = val;
      }
      if (testcase.error) {
        testcase.error = testcase.error.message;
      }
      debug('ran test', '"testcase.name"', testcase.passed, err);
      return done(null); // ignore error to not bail out early
    });
  };
  runOneSuite = function runOneSuite(suite, cb) {
    return runSuite(runner, suite, runTest, cb);
  };
  debug('running suites', function () {
    var j, len, results1;
    results1 = [];
    for (j = 0, len = suites.length; j < len; j++) {
      s = suites[j];
      results1.push(s.name);
    }
    return results1;
  }());
  return common.asyncSeries(suites, runOneSuite, function (err) {
    return doneCallback(err);
  });
};

exports.Runner = Runner;

exports.runAll = runAll;

exports.runTestAndCheck = runTestAndCheck;